import datetime

from toga_web.libs import create_proxy

from .base import Widget


def py_time(native_time):
    if isinstance(native_time, datetime.datetime):
        time = native_time.time()
    elif isinstance(native_time, str):
        time = datetime.time.fromisoformat(native_time)
    else:
        time = datetime.datetime.now().time()

    return time.replace(microsecond=0)


def native_time(py_time):
    if isinstance(py_time, datetime.datetime):
        time = py_time.time()
    else:
        time = py_time

    return time.replace(microsecond=0).isoformat()


class TimeInput(Widget):

    def create(self):
        self.native = self._create_native_widget("sl-input")
        self.native.setAttribute("step", "1")  # force seconds to show
        self.native.type = "time"
        self.native.value = native_time(datetime.datetime.now())
        self.native.addEventListener("sl-change", create_proxy(self.dom_sl_change))

    def dom_sl_change(self, event):
        try:
            input_time = py_time(self.native.value)
        except Exception:
            input_time = datetime.datetime.now().time().replace(microsecond=0)

        self.set_value(input_time)
        self.interface.on_change()

    def get_value(self):
        try:
            return py_time(self.native.value)
        except Exception:
            return datetime.datetime.now().time().replace(microsecond=0)

    def set_value(self, value):
        if value is None or not isinstance(value, datetime.time):
            value = datetime.datetime.now().time().replace(microsecond=0)

        min_time = self.get_min_time()
        max_time = self.get_max_time()

        if value < min_time:
            value = min_time
        elif value > max_time:
            value = max_time
        self.native.value = native_time(value)

    def get_min_time(self):
        min_value = self.native.min if self.native.min else datetime.time(0, 0, 0)
        return py_time(min_value)

    def set_min_time(self, value):
        self.native.min = native_time(value)

    def get_max_time(self):
        max_value = self.native.max if self.native.max else datetime.time(23, 59, 59)
        return py_time(max_value)

    def set_max_time(self, value):
        self.native.max = native_time(value)



from .base import Widget


class ProgressBar(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-progress-bar")
        self._is_running = False
        self._value = 0
        self._max = 0

    def is_running(self):
        return self._is_running

    def get_value(self):
        return self._value

    def set_value(self, value):
        self._value = value
        self.native.value = self.percentage

    def get_max(self):
        return self._max

    def set_max(self, value):
        self._max = value
        if value is None:
            if self._is_running:
                self.native.indeterminate = True
        else:
            self.native.value = self.percentage
            self.native.indeterminate = False

    def start(self):
        self._is_running = True
        if self._max is None:
            self.native.indeterminate = True

    def stop(self):
        self._is_running = False
        self.native.indeterminate = False

    @property
    def percentage(self):
        if self._max is None or self._max == 0:
            return 0

        return self._value / self._max * 100



from toga_web.libs import create_proxy

from .base import Widget


class Switch(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-switch")
        self.native.addEventListener("sl-change", create_proxy(self.dom_onchange))

    def dom_onchange(self, event):
        self.interface.on_change()

    def get_text(self):
        return self.native.innerHTML

    def set_text(self, text):
        self.native.innerHTML = text

    def get_value(self):
        return self.native.checked

    def set_value(self, value):
        old_value = self.get_value()
        self.native.checked = value
        if value != old_value:
            self.interface.on_change()



from toga.constants import Direction

from .base import Widget


class Divider(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-divider")

    def get_direction(self):
        return self.interface.direction

    def set_direction(self, value):
        if value is Direction.VERTICAL:
            self.native.setAttribute("vertical", "")
        else:
            self.native.removeAttribute("vertical")



from toga_web.libs import create_proxy

from .base import Widget


class Button(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-button")
        self.native.addEventListener("onclick", create_proxy(self.dom_onclick))

    def dom_onclick(self, event):
        self.interface.on_press()

    def get_text(self):
        return self.native.innerHTML

    def set_text(self, text):
        self.native.innerHTML = text

    def get_icon(self):
        return self._icon

    def set_icon(self, icon):
        self._icon = icon
        if icon:
            self.interface.factory.not_implemented("Button.icon")

    def set_enabled(self, value):
        self.native.disabled = not value

    def set_background_color(self, value):
        pass

    def rehint(self):
        pass



from abc import ABC

import toga
from toga_web.libs import create_element


class BaseDialog(ABC):
    def show(self, host_window, future):
        self.future = future

        if self.native:
            # Add the dialog to the DOM. Don't differentiate between app and window
            # modal dialogs.
            toga.App.app._impl.native.appendChild(self.native)

            self.native.show()
        else:
            # Dialog doesn't have an implementation
            self.future.set_result(None)


class InfoDialog(BaseDialog):
    def __init__(self, title, message):
        super().__init__()
        self.native = create_element(
            "sl-dialog",
            id="toga-info-dialog",
            label=title,
            children=[
                create_element("p", content=message),
            ]
            + self.create_buttons(),
        )

    def create_buttons(self):
        close_button = create_element(
            "sl-button", slot="footer", variant="primary", content="Ok"
        )
        # Handle the close of the dialog
        close_button.onclick = self.dialog_close

        return [close_button]

    def dialog_close(self, event):
        self.native.hide()
        self.native.parentElement.removeChild(self.native)

        self.future.set_result(None)


class QuestionDialog(BaseDialog):
    def __init__(self, title, message):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.QuestionDialog()")
        self.native = None


class ConfirmDialog(BaseDialog):
    def __init__(self, title, message):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.ConfirmDialog()")
        self.native = None


class ErrorDialog(BaseDialog):
    def __init__(self, title, message):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.ErrorDialog()")
        self.native = None


class StackTraceDialog(BaseDialog):
    def __init__(self, title, message, **kwargs):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.StackTraceDialog()")
        self.native = None


class SaveFileDialog(BaseDialog):
    def __init__(
        self,
        interface,
        title,
        filename,
        initial_directory,
        file_types=None,
    ):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.SaveFileDialog()")
        self.native = None


class OpenFileDialog(BaseDialog):
    def __init__(
        self,
        interface,
        title,
        initial_directory,
        file_types,
        multiselect,
    ):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.OpenFileDialog()")
        self.native = None


class SelectFolderDialog(BaseDialog):
    def __init__(
        self,
        interface,
        title,
        initial_directory,
        multiselect,
    ):
        super().__init__()

        toga.App.app.factory.not_implemented("dialogs.SelectFolderDialog()")
        self.native = None



from pathlib import Path


class Paths:
    def __init__(self, interface):
        self.interface = interface

    def get_config_path(self):
        return Path.home() / "config"

    def get_data_path(self):
        return Path.home() / "data"

    def get_cache_path(self):
        return Path.home() / "cache"

    def get_logs_path(self):
        return Path.home() / "log"



from .base import Widget


class ActivityIndicator(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-spinner")
        self.stop()

    # Actions
    def start(self):
        self.native.style.visibility = "visible"
        self._is_running = True

    def stop(self):
        self.native.style.visibility = "hidden"
        self._is_running = False

    def is_running(self):
        return self._is_running



import asyncio

import toga
from toga_web.libs import create_element, js

from .screens import Screen as ScreenImpl


class App:
    # Web apps exit when the last window is closed
    CLOSE_ON_LAST_WINDOW = True
    # Web apps use default command line handling
    HANDLES_COMMAND_LINE = False

    def __init__(self, interface):
        self.interface = interface
        self.interface._impl = self

        self.loop = asyncio.new_event_loop()

    def create(self):
        self.native = js.document.getElementById("app-placeholder")
        # self.resource_path = ???

        # Call user code to populate the main window
        self.interface._startup()

    ######################################################################
    # Commands and menus
    ######################################################################

    def create_standard_commands(self):
        pass

    def create_menus(self):
        # Web menus are created on the Window.
        for window in self.interface.windows:
            # It's difficult to trigger this on a simple window, because we can't easily
            # modify the set of app-level commands that are registered, and a simple
            # window doesn't exist when the app starts up. Therefore, no-branch the else
            # case.
            if hasattr(window._impl, "create_menus"):  # pragma: no branch
                window._impl.create_menus()

    ######################################################################
    # App lifecycle
    ######################################################################

    def exit(self):
        pass

    def main_loop(self):
        self.create()

    def set_icon(self, icon):
        pass

    def set_main_window(self, window):
        if window is None:
            raise RuntimeError("Session-based apps are not supported on Web")
        elif window == toga.App.BACKGROUND:
            raise RuntimeError("Background apps are not supported on Web")

    ######################################################################
    # App resources
    ######################################################################

    def get_screens(self):
        return [ScreenImpl(js.document.documentElement)]

    ######################################################################
    # App state
    ######################################################################

    def get_dark_mode_state(self):
        self.interface.factory.not_implemented("dark mode state")
        return None

    ######################################################################
    # App capabilities
    ######################################################################

    def beep(self):
        self.interface.factory.not_implemented("App.beep()")

    def show_about_dialog(self):
        name_and_version = f"{self.interface.formal_name}"

        if self.interface.version is not None:
            name_and_version += f" v{self.interface.version}"

        if self.interface.author is not None:
            copyright = f"\n\nCopyright © {self.interface.author}"

        close_button = create_element(
            "sl-button", slot="footer", variant="primary", content="Ok"
        )
        about_dialog = create_element(
            "sl-dialog",
            id="toga-about-dialog",
            label="About",
            children=[
                create_element("p", content=name_and_version),
                create_element("p", content=copyright),
                close_button,
            ],
        )

        # Create a button handler to capture the close,
        # and destroy the dialog
        def dialog_close(event):
            about_dialog.hide()
            self.native.removeChild(about_dialog)

        close_button.onclick = dialog_close

        # Add the dialog to the DOM.
        self.native.appendChild(about_dialog)

        # If this is the first time a dialog is being shown, the Shoelace
        # autoloader needs to construct the Dialog custom element. We can't
        # display the dialog until that element has been fully loaded and
        # constructed. Only show the dialog when the promise of <sl-dialog>
        # element construction has been fulfilled.
        def show_dialog(promise):
            about_dialog.show()

        js.customElements.whenDefined("sl-dialog").then(show_dialog)

    ######################################################################
    # Cursor control
    ######################################################################

    def show_cursor(self):
        self.interface.factory.not_implemented("App.show_cursor()")

    def hide_cursor(self):
        self.interface.factory.not_implemented("App.hide_cursor()")

    ######################################################################
    # Window control
    ######################################################################

    def get_current_window(self):
        self.interface.factory.not_implemented("App.get_current_window()")

    def set_current_window(self):
        self.interface.factory.not_implemented("App.set_current_window()")

    ######################################################################
    # Presentation mode controls
    ######################################################################

    def enter_presentation_mode(self, screen_window_dict):
        self.interface.factory.not_implemented("App.enter_presentation_mode()")

    def exit_presentation_mode(self):
        self.interface.factory.not_implemented("App.exit_presentation_mode()")



from toga_web.libs import create_element


class Widget:
    def __init__(self, interface):
        self.interface = interface
        self._container = None

        self.create()

    def _create_native_widget(
        self, tag, classes=None, content=None, children=None, **properties
    ):
        """Create a DOM element representing a native widget.

        The ID and style of the widget will be automatically set;
        ``toga``, and the name of the widget class (in lower case)
        will be added as a class name on the widget.

        :param widget: The web implementation being created.
        :param tag: The HTML tag for t
        :param classes: (Optional) A list of classes to attach to the
            new element. Two widgets
        :param content: (Optional) The innerHTML content of the element.
        :param children: (Optional) A list of direct descendents to add to
            the element.
        :param properties: Any additional properties that should be set.
            These *must* be HTML DOM properties (e.g., ``readOnly``);
            they cannot be events or methods.
        :returns: A newly created DOM element.
        """
        if classes is None:
            classes = []

        classes = ["toga", self.interface.__class__.__name__.lower()] + classes

        native = create_element(
            tag,
            id=f"toga_{self.interface.id}",
            classes=classes,
            style=self.interface.style.__css__(),
            content=content,
            children=children,
            **properties,
        )

        return native

    def create(self):
        raise NotImplementedError()

    def set_app(self, app):
        pass

    def set_window(self, window):
        pass

    @property
    def viewport(self):
        return self._container

    @property
    def container(self):
        return self._container

    @container.setter
    def container(self, container):
        self._container = container

        for child in self.interface.children:
            child._impl.container = container

    def get_enabled(self):
        return not self.native.disabled

    def set_enabled(self, value):
        self.native.disabled = not value

    def focus(self):
        self.interface.factory.not_implemented("Widget.focus()")

    def get_tab_index(self):
        self.interface.factory.not_implementated("Widget.get_tab_index()")

    def set_tab_index(self, tab_index):
        self.interface.factory.not_implementated("Widget.set_tab_index()")

    ######################################################################
    # APPLICATOR
    #
    # Web style is a little different to other platforms; we if there's
    # any change, we can just re-set the CSS styles and the browser
    # will reflect those changes as needed.
    ######################################################################

    def _reapply_style(self):
        self.native.style = self.interface.style.__css__()

    def set_bounds(self, x, y, width, height):
        self._reapply_style()

    def set_text_align(self, alignment):
        self._reapply_style()

    def set_hidden(self, hidden):
        self._reapply_style()

    def set_font(self, font):
        self._reapply_style()

    def set_color(self, color):
        self._reapply_style()

    def set_background_color(self, color):
        self._reapply_style()

    ######################################################################
    # INTERFACE
    ######################################################################

    def add_child(self, child):
        pass

    def insert_child(self, index, child):
        self.add_child(child)

    def remove_child(self, child):
        child.container = None

    def refresh(self):
        self._reapply_style()

    def rehint(self):
        pass



from toga import Command as StandardCommand, Group


class Command:
    """Command `native` property is a list of native widgets associated with the
    command."""

    def __init__(self, interface):
        self.interface = interface
        self.native = []

    @classmethod
    def standard(cls, app, id):
        # ---- Help menu ----------------------------------
        if id == StandardCommand.ABOUT:
            return {
                "text": f"About {app.formal_name}",
                "group": Group.HELP,
            }
        elif id == StandardCommand.PREFERENCES:
            return {
                "text": "Preferences",
                "group": Group.HELP,
            }
        # ---- Non-existent commands ----------------------------------
        elif id in {
            StandardCommand.EXIT,
            StandardCommand.NEW,
            StandardCommand.OPEN,
            StandardCommand.SAVE,
            StandardCommand.SAVE_AS,
            StandardCommand.SAVE_ALL,
            StandardCommand.VISIT_HOMEPAGE,
        }:
            # These commands are valid, but don't exist on web.
            return None

        raise ValueError(f"Unknown standard command {id!r}")

    def dom_click(self, event):
        self.interface.action()

    def set_enabled(self, value):
        pass
        # enabled = self.interface.enabled
        # for widget in self.native:
        #     try:
        #         widget.set_sensitive(enabled)
        #     except AttributeError:
        #         widget.set_enabled(enabled)



from .base import Widget


class Label(Widget):
    def create(self):
        self.native = self._create_native_widget("span")

    def get_text(self):
        return self.native.innerHTML

    def set_text(self, value):
        self.native.innerHTML = value

    def set_text_align(self, value):
        pass

    def rehint(self):
        pass



from .textinput import TextInput


class PasswordInput(TextInput):
    def create(self):
        super().create()
        self.native.setAttribute("type", "password")

    def is_valid(self):
        self.interface.factory.not_implemented("PasswordInput.is_valid()")
        return True



from toga.command import Group, Separator
from toga.constants import WindowState
from toga.types import Position, Size
from toga_web.libs import create_element, create_proxy, js

from .screens import Screen as ScreenImpl


class Window:
    def __init__(self, interface, title, position, size):
        self.interface = interface
        self.interface._impl = self

        self.native = create_element(
            "main",
            id=f"toga_{self.interface.id}",
            classes=["toga", "window", "container"],
            role="main",
        )

        app_placeholder = js.document.getElementById("app-placeholder")
        app_placeholder.appendChild(self.native)

        js.document.body.addEventListener(
            "focus", create_proxy(self.dom_on_gain_focus), True
        )
        js.document.body.addEventListener(
            "blur", create_proxy(self.dom_on_lose_focus), True
        )
        js.document.addEventListener(
            "visibilitychange", create_proxy(self.dom_on_visibility_change)
        )

        self.set_title(title)

    ######################################################################
    # Native event handlers
    ######################################################################

    def on_close(self, *args):
        pass

    def on_size_allocate(self, widget, allocation):
        pass

    def dom_on_gain_focus(self, event):
        self.interface.on_gain_focus()

    def dom_on_lose_focus(self, event):
        self.interface.on_lose_focus()

    def dom_on_visibility_change(self, event):
        if js.document.visibilityState == "visible":
            self.interface.on_show()
        else:
            self.interface.on_hide()

    ######################################################################
    # Window properties
    ######################################################################

    def get_title(self):
        return js.document.title

    def set_title(self, title):
        js.document.title = title

    ######################################################################
    # Window lifecycle
    ######################################################################

    def close(self):
        self.interface.factory.not_implemented("Window.close()")

    def set_app(self, app):
        pass

    def show(self):
        self.native.style = "visibility: visible;"

    ######################################################################
    # Window content and resources
    ######################################################################

    def clear_content(self):
        if self.interface.content:
            for child in self.interface.content.children:
                child._impl.container = None

    def set_content(self, widget):
        # Remove existing content of the window.
        for child in self.native.childNodes:
            self.native.removeChild(child)

        # Add all children to the content widget.
        self.native.appendChild(widget.native)

    ######################################################################
    # Window size
    ######################################################################

    def get_size(self) -> Size:
        return Size(self.native.offsetWidth, self.native.offsetHeight)

    def set_size(self, size):
        # Does nothing on web
        pass

    ######################################################################
    # Window position
    ######################################################################

    def get_current_screen(self):
        return ScreenImpl(js.document.documentElement)

    def get_position(self) -> Position:
        return Position(0, 0)

    def set_position(self, position):
        # Does nothing on web
        pass

    ######################################################################
    # Window visibility
    ######################################################################

    def get_visible(self):
        self.interface.factory.not_implemented("Window.get_visible()")

    def hide(self):
        self.native.style = "visibility: hidden;"

    ######################################################################
    # Window state
    ######################################################################

    def get_window_state(self, in_progress_state=False):
        # Windows are always normal
        return WindowState.NORMAL

    def set_window_state(self, state):
        self.interface.factory.not_implemented("Window.set_window_state()")

    ######################################################################
    # Window capabilities
    ######################################################################

    def get_image_data(self):
        self.interface.factory.not_implemented("Window.get_image_data()")


class MainWindow(Window):
    def _create_submenu(self, group, items):
        submenu = create_element(
            "sl-dropdown",
            children=[
                create_element(
                    "span",
                    id=f"menu-{id(group)}",
                    classes=["menu"],
                    slot="trigger",
                    content=group.text,
                ),
                create_element(
                    "sl-menu",
                    children=items,
                ),
            ],
        )
        return submenu

    def create_menus(self):
        self._menu_groups = {}
        submenu = None

        for cmd in self.interface.app.commands:
            if isinstance(cmd, Separator):
                # TODO - add a section break
                pass
            else:
                # TODO - this doesn't handle submenus properly;
                # all menu items will appear in their root group.
                submenu = self._menu_groups.setdefault(cmd.group, [])

                menu_item = create_element(
                    "sl-menu-item",
                    content=cmd.text,
                    disabled=not cmd.enabled,
                )
                menu_item.onclick = cmd._impl.dom_click

                submenu.append(menu_item)

        menu_container = create_element("nav", classes=["menubar"])
        help_menu_container = create_element("nav", classes=["menubar"])

        # If there isn't an explicit app menu group, add an inert placeholder
        if Group.APP not in self._menu_groups:
            menu_container.appendChild(
                create_element(
                    "span",
                    classes=["app"],
                    content=self.interface.app.formal_name,
                )
            )

        for group, items in self._menu_groups.items():
            submenu = self._create_submenu(group, items)
            if group != Group.HELP:
                menu_container.appendChild(submenu)
            else:
                help_menu_container.appendChild(submenu)

        menubar_id = f"{self.interface.id}-header"
        self.menubar = create_element(
            "header",
            id=menubar_id,
            classes=["toga"],
            children=[
                create_element(
                    "a",
                    classes=["brand"],
                    children=[
                        create_element(
                            "img",
                            src="static/logo-32.png",
                            alt=f"{self.interface.app.formal_name} logo",
                            loading="lazy",
                        )
                    ],
                ),
                menu_container,
                help_menu_container,
            ],
        )

        # If there's an existing menubar, replace it.
        old_menubar = js.document.getElementById(menubar_id)
        if old_menubar:
            old_menubar.replaceWith(self.menubar)
        else:
            self.native.prepend(self.menubar)

    def create_toolbar(self):
        self.interface.factory.not_implemented("Window.create_toolbar()")



from .base import Widget


class Box(Widget):
    def create(self):
        self.native = self._create_native_widget("div", classes=["container"])

    def add_child(self, child):
        self.native.appendChild(child.native)



import toga


class StatusIcon:
    def __init__(self, interface):
        self.interface = interface
        self.native = None

    def set_icon(self, icon):
        pass

    def create(self):
        toga.NotImplementedWarning.warn("Web", "Status Icons")

    def remove(self):
        pass


class SimpleStatusIcon(StatusIcon):
    pass


class MenuStatusIcon(StatusIcon):
    pass


class StatusIconSet:
    def __init__(self, interface):
        self.interface = interface

    def create(self):
        pass



from toga.screens import Screen as ScreenInterface
from toga.types import Position, Size


class Screen:
    _instances = {}

    def __new__(cls, native):
        if native in cls._instances:
            return cls._instances[native]
        else:
            instance = super().__new__(cls)
            instance.interface = ScreenInterface(_impl=instance)
            instance.native = native
            cls._instances[native] = instance
            return instance

    def get_name(self):
        return "Web Screen"

    def get_origin(self) -> Position:
        return Position(0, 0)

    def get_size(self) -> Size:
        return Size(self.native.clientWidth, self.native.clientHeight)

    def get_image_data(self):
        self.interface.factory.not_implemented("Screen.get_image_data()")



import datetime

from toga_web.libs import create_proxy

from .base import Widget


def py_date(native_value):
    return datetime.date.fromisoformat(native_value)


def native_date(py_date_obj):
    return py_date_obj.isoformat()


class DateInput(Widget):
    def create(self):
        self.native = self._create_native_widget("sl-input")
        self.native.type = "date"
        self.native.value = native_date(datetime.date.today())
        self.native.addEventListener("sl-change", create_proxy(self.dom_sl_change))

    def dom_sl_change(self, event):
        try:
            input_date = py_date(self.native.value)
        except Exception:
            input_date = datetime.date.today()

        self.set_value(input_date)
        self.interface.on_change()

    def get_value(self):
        try:
            return py_date(self.native.value)
        except Exception:
            return datetime.date.today()

    def set_value(self, value):
        if value is None:
            value = datetime.date.today()

        min_date = self.get_min_date()
        max_date = self.get_max_date()

        if value < min_date:
            value = min_date
        elif value > max_date:
            value = max_date

        self.native.value = native_date(value)

    def set_min_date(self, value):
        self.native.min = native_date(value)

    def get_min_date(self):
        if self.native.min is None:
            return datetime.date(1800, 1, 1)
        return datetime.date.fromisoformat(str(self.native.min))

    def set_max_date(self, value):
        self.native.max = native_date(value)

    def get_max_date(self):
        if self.native.max is None:
            return datetime.date(8999, 12, 31)
        return datetime.date.fromisoformat(str(self.native.max))



from toga_web.libs import create_proxy

from .base import Widget


class ScrollContainer(Widget):
    def __init__(self, interface):
        super().__init__(interface)
        self.native.addEventListener("scroll", create_proxy(self.dom_onscroll))

    def create(self):
        self._horizontal_enabled = True
        self._vertical_enabled = True

        self.native = self._create_native_widget(
            "div",
            classes=["container"],
        )
        self._content_div = self._create_native_widget(
            "div",
            classes=["toga-scroll-content"],
        )
        self.native.appendChild(self._content_div)
        self._update_overflow()

    def set_content(self, content_impl):
        while self._content_div.firstChild:
            self._content_div.removeChild(self._content_div.firstChild)

        if content_impl is not None:
            self._content_div.appendChild(content_impl.native)

    def _update_overflow(self):
        self.native.style.overflowX = "auto" if self._horizontal_enabled else "hidden"
        self.native.style.overflowY = "auto" if self._vertical_enabled else "hidden"

    def get_horizontal(self):
        return self._horizontal_enabled

    def set_horizontal(self, value):
        self._horizontal_enabled = bool(value)
        self._update_overflow()

    def get_vertical(self):
        return self._vertical_enabled

    def set_vertical(self, value):
        self._vertical_enabled = bool(value)
        self._update_overflow()

    def get_max_horizontal_position(self):
        return max(0, self.native.scrollWidth - self.native.clientWidth)

    def get_max_vertical_position(self):
        return max(0, self.native.scrollHeight - self.native.clientHeight)

    def get_horizontal_position(self):
        return self.native.scrollLeft

    def get_vertical_position(self):
        return self.native.scrollTop

    def set_position(self, horizontal, vertical):
        if self._horizontal_enabled:
            horizontal = max(
                0, min(int(horizontal), self.get_max_horizontal_position())
            )
        else:
            horizontal = self.native.scrollLeft

        if self._vertical_enabled:
            vertical = max(0, min(int(vertical), self.get_max_vertical_position()))
        else:
            vertical = self.native.scrollTop

        self.native.scrollTo(horizontal, vertical)

    def dom_onscroll(self, _evt):
        self.interface.on_scroll()



try:
    # Try to import js from the PyScript namespace
    import js
except ModuleNotFoundError:
    # To ensure the code can be imported, provide a js symbol as a fallback
    js = None
try:
    import pyodide
except ModuleNotFoundError:
    pyodide = None

from pyscript.web import document

create_proxy = pyodide.ffi.create_proxy if pyodide else lambda f: f


def create_element(
    tag,
    id=None,
    classes=None,
    style=None,
    disabled=False,
    content=None,
    children=None,
    **properties,
):
    """Utility method for creating DOM elements.

    :param tag: The HTML tag of the element to create
    :param id: (Optional) The ID of the new element
    :param classes: (Optional) A list of classes to attach to the new element.
    :param style: (Optional) The CSS style for the element.
    :param disabled: (Optional) Should the element be disabled at time of
        creation? (Default: False)
    :param content: (Optional) The innerHTML content of the element.
    :param children: (Optional) A list of direct descendents to add to the
        element.
    :param properties: Any additional properties that should be set. These
        *must* be HTML DOM properties (e.g., ``readOnly``); they cannot be
        events or methods.
    :returns: A newly created DOM element.
    """
    element = document.createElement(tag)

    if id:
        element.id = id

    if classes:
        for klass in classes:
            element.classList.add(klass)

    if style:
        element.style = style

    if disabled:
        element.setAttribute("disabled", "")

    for attr, value in properties.items():
        element.setAttribute(attr.replace("_", "-"), value)

    if content:
        element.innerHTML = content

    if children:
        for child in children:
            element.appendChild(child)

    return element



from toga import NotImplementedWarning

from . import dialogs
from .app import App
from .command import Command
from .fonts import Font
from .icons import Icon

# from .images import Image
from .paths import Paths
from .statusicons import MenuStatusIcon, SimpleStatusIcon, StatusIconSet
from .widgets.activityindicator import ActivityIndicator
from .widgets.box import Box
from .widgets.button import Button
from .widgets.dateinput import DateInput
from .widgets.divider import Divider

# from .widgets.canvas import Canvas
# from .widgets.detailedlist import DetailedList
# from .widgets.imageview import ImageView
from .widgets.label import Label

# from .widgets.multilinetextinput import MultilineTextInput
# from .widgets.numberinput import NumberInput
# from .widgets.optioncontainer import OptionContainer
from .widgets.passwordinput import PasswordInput
from .widgets.progressbar import ProgressBar
from .widgets.scrollcontainer import ScrollContainer

# from .widgets.selection import Selection
# from .widgets.slider import Slider
# from .widgets.splitcontainer import SplitContainer
from .widgets.switch import Switch

# from .widgets.table import Table
from .widgets.textinput import TextInput
from .widgets.timeinput import TimeInput

# from .widgets.tree import Tree
# from .widgets.webview import WebView
from .window import MainWindow, Window


def not_implemented(feature):
    NotImplementedWarning.warn("Web", feature)  # pragma: nocover


__all__ = [
    "not_implemented",
    "App",
    "Command",
    # Resources
    "Font",
    "Icon",
    # 'Image',
    "Paths",
    "dialogs",
    # Status Icons
    "MenuStatusIcon",
    "SimpleStatusIcon",
    "StatusIconSet",
    # # Widgets
    "Box",
    "Button",
    # 'Canvas',
    "Divider",
    "DateInput",
    "TimeInput",
    # 'DetailedList',
    # 'ImageView',
    "Label",
    # 'MultilineTextInput',
    # 'NumberInput',
    # 'OptionContainer',
    "PasswordInput",
    "ProgressBar",
    "ActivityIndicator",
    "ScrollContainer",
    # 'Selection',
    # 'Slider',
    # 'SplitContainer',
    "Switch",
    # 'Table',
    "TextInput",
    # 'Tree',
    # 'WebView',
    # Windows
    "MainWindow",
    "Window",
]


def __getattr__(name):  # pragma: no cover
    raise NotImplementedError(f"Toga's Web backend doesn't implement {name}")



from toga_web.libs import create_proxy

from .base import Widget


class TextInput(Widget):
    def create(self):
        self._return_listener = None
        self.native = self._create_native_widget("sl-input")
        self.native.onkeyup = self.dom_onkeyup
        self.native.addEventListener("onkeyup", create_proxy(self.dom_onkeyup))
        self.native.addEventListener("sl-change", create_proxy(self.dom_sl_change))

    def dom_onkeyup(self, event):
        self.interface.on_change()

    def dom_sl_change(self, event):
        self.interface.on_confirm()

    def set_readonly(self, value):
        self.native.readOnly = value

    def set_placeholder(self, value):
        if value:
            self.native.placeholder = value

    def get_value(self):
        return self.native.value

    def set_value(self, value):
        self.native.value = value

    def set_font(self, font):
        pass

    def set_text_align(self, value):
        pass

    def rehint(self):
        pass

    def set_on_change(self, handler):
        pass

    def set_on_gain_focus(self, handler):
        pass

    def set_on_lose_focus(self, handler):
        pass

    def set_error(self, error_message):
        pass

    def clear_error(self):
        pass

    def is_valid(self):
        self.interface.factory.not_implemented("TextInput.is_valid()")
        return True
